@@ -1,5 +1,5 @@
 /* Fault handler information.  Unix version.
-   Copyright (C) 1993-1999, 2002  Bruno Haible <bruno@clisp.org>
+   Copyright (C) 1993-1999, 2002-2003  Bruno Haible <bruno@clisp.org>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -54,8 +54,16 @@
 #include <stddef.h> /* needed for NULL on SunOS4 */
 #include <stdlib.h>
 #include <signal.h>
+#if HAVE_SYS_SIGNAL_H
+# include <sys/signal.h>
+#endif
 #include <errno.h>
 
+/* For MacOSX.  */
+#ifndef SS_DISABLE
+#define SS_DISABLE SA_DISABLE
+#endif
+
 #include "fault.h"
 #include CFG_SIGNALS
 
@@ -77,7 +85,7 @@
    Leaving a signal handler executing on the alternate stack.  */
 #include "leave.h"
 
-#ifdef CFG_STACKVMA
+#if HAVE_STACKVMA
 
 /* Address of the last byte belonging to the stack vma.  */
 static unsigned long stack_top = 0;
@@ -92,7 +100,7 @@ remember_stack_top (void *some_variable_on_stack)
     stack_top = vma.end - 1;
 }
 
-#endif /* CFG_STACKVMA */
+#endif /* HAVE_STACKVMA */
 
 static stackoverflow_handler_t stk_user_handler = (stackoverflow_handler_t)NULL;
 static unsigned long stk_extra_stack;
@@ -118,8 +126,8 @@ sigsegv_handler (SIGSEGV_FAULT_HANDLER_ARGLIST)
   void *address = (void *) (SIGSEGV_FAULT_ADDRESS);
 
 #if HAVE_STACK_OVERFLOW_RECOVERY
-#if !(defined CFG_STACKVMA || defined SIGSEGV_FAULT_STACKPOINTER)
-#error "Insufficient heuristics for detecting a stack overflow.  Either define CFG_STACKVMA correctly, or define SIGSEGV_FAULT_STACKPOINTER correctly, or undefine HAVE_STACK_OVERFLOW_RECOVERY!"
+#if !(HAVE_STACKVMA || defined SIGSEGV_FAULT_STACKPOINTER)
+#error "Insufficient heuristics for detecting a stack overflow.  Either define CFG_STACKVMA and HAVE_STACKVMA correctly, or define SIGSEGV_FAULT_STACKPOINTER correctly, or undefine HAVE_STACK_OVERFLOW_RECOVERY!"
 #endif
 
   /* Call user's handler.  */
@@ -137,9 +145,12 @@ sigsegv_handler (SIGSEGV_FAULT_HANDLER_ARGLIST)
           /* See whether it was a stack overflow. If so, longjump away.  */
 #ifdef SIGSEGV_FAULT_STACKPOINTER
           unsigned long old_sp = (unsigned long) (SIGSEGV_FAULT_STACKPOINTER);
+#ifdef __ia64
+          unsigned long old_bsp = (unsigned long) (SIGSEGV_FAULT_BSP_POINTER);
+#endif
 #endif
 
-#ifdef CFG_STACKVMA
+#if HAVE_STACKVMA
           /* Were we able to determine the stack top?  */
           if (stack_top)
             {
@@ -150,9 +161,15 @@ sigsegv_handler (SIGSEGV_FAULT_HANDLER_ARGLIST)
                 {
                   /* Heuristic AC: If the fault_address is nearer to the stack
                      segment's [start,end] than to the previous segment, we
-                     consider it a stack overflow.  */
+                     consider it a stack overflow.
+                     In the case of IA-64, we know that the previous segment
+                     is the up-growing bsp segment, and either of the two
+                     stacks can overflow.  */
                   unsigned long addr = (unsigned long) address;
 
+#ifdef __ia64
+                  if (addr >= vma.prev_end && addr <= vma.end - 1)
+#else
 #if STACK_DIRECTION < 0
                   if (addr >= vma.start
                       ? (addr <= vma.end - 1)
@@ -162,12 +179,17 @@ sigsegv_handler (SIGSEGV_FAULT_HANDLER_ARGLIST)
                       ? (addr >= vma.start)
                       : (addr - vma.end < (vma.next_start - vma.end) / 2))
 #endif
+#endif
 #else
           /* Heuristic AB: If the fault address is near the stack pointer,
              it's a stack overflow.  */
           unsigned long addr = (unsigned long) address;
 
-          if (addr <= old_sp + 4096 && old_sp <= addr + 4096)
+          if ((addr <= old_sp + 4096 && old_sp <= addr + 4096)
+#ifdef __ia64
+              || (addr <= old_bsp + 4096 && old_bsp <= addr + 4096)
+#endif
+             )
             {
                 {
 #endif
@@ -435,7 +457,7 @@ stackoverflow_install_handler (stackoverflow_handler_t handler,
                                void *extra_stack, unsigned long extra_stack_size)
 {
 #if HAVE_STACK_OVERFLOW_RECOVERY
-#ifdef CFG_STACKVMA
+#if HAVE_STACKVMA
   if (!stack_top)
     {
       int dummy;
